[![Build status](https://github.com/Armavica/rebop/actions/workflows/rust.yml/badge.svg)](https://github.com/Armavica/rebop/actions/)
[![Crates.io](https://img.shields.io/crates/v/rebop)](https://crates.io/crates/rebop/)
[![Docs.rs](https://docs.rs/rebop/badge.svg)](https://docs.rs/rebop/)

# {{crate}}

{{readme}}

License: MIT
